import math

from players.entity import PlayerEntity
from players.helpers import userid_from_index
from players.helpers import index_from_userid

from entities.entity import BaseEntity
from entities.helpers import inthandle_from_index

from filters.players import PlayerIter
from filters.entities import EntityIter
from filters.recipients import RecipientFilter

from messages import SayText2
from messages import HintText

from engines.server import engine_server
from engines.sound import Sound, VOL_NORM, SOUND_FROM_WORLD
from engines.trace import engine_trace
from engines.trace import Ray
from engines.trace import GameTrace
from engines.trace import TraceType
from engines.trace import TraceFilterSimple

from mathlib import Vector

from effects import temp_entities

from events import Event




# we don't need to get the PlayerEntity instance for every index everytime
# so we just save them to use the same instance again
_players = {}

# we don't want to precache a model everytime
# so we just save their indexes to use them again
_precached_models = {}


@Event
def game_start(game_event):
    # Clear the dictionaries
    _players.clear()
    _precached_models.clear()


def getindex(userid):
    # Returns the index of a player by its userid

    return index_from_userid(userid)

def getuserid(index):
    # Returns the userid of a player by its index

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    return player.userid

def getplayername(index):
    # Returns the name of a player

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    return player.name

def getplayersteamid(index):
    # Returns the steamid of a player

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    return player.steamid

def getplayerpointer(index):
    # Returns the pointer of a player

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    return player.pointer

def getplayerinfo(index):
    # Returns the playerinfo of a player

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    return player.playerinfo

def getplayerhandle(index):
    # Returns the handle of a player

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    return player.inthandle

def getplayerlocation(index):
    # Returns a mathlib.Vector instance with the player's location

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    return player.location

def getplayerteam(index):
    # Returns the team of a player (0, 1, 2, 3)

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    return player.team

def gethealth(index):
    # Returns the health of a player

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    return player.health

def getarmor(index):
    # Returns the armor of a player

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    return player.armor

def getmodel(index):
    # Returns the model name of a player

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    return player.model

def getcolor(index):
    # Returns the model color of a player in RGBA format

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    color = player.get_prop_int("m_clrRender")
    return tuple(int(x) for x in (color & 0xff, (color & 0xff00) >> 8, (color & 0xff0000) >> 16, (color & 0xff000000) >> 24))

def getlanguage(index):
    # Returns the cl_language value of a player

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    return player.language

def isdead(index):
    # Returns True if the player is dead

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    return player.isdead

def iscrouched(index):
    # Returns True if the player is crouched

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    return True if player.get_prop_int('localdata.m_Local.m_bDucked') else False

def getviewcoordinates(index, ignore=(), trace_type=TraceType.EVERYTHING):
    # Returns the coordinates a player is currently looking at.

    """
    http://www.sourcepython.com/showthread.php?579-Player-View-Coordinates

    @param <ignore>:
    An iterable of entity indexes to ignore. The trace will not hit these
    entities.

    @param <trace_type>:
    Defines the trace type.

    """

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    start_vec = geteyelocation(index)
    end_vec = start_vec + getviewvector(index) * 16384
    trace = GameTrace()
    engine_trace.trace_ray(
        Ray(start_vec, end_vec), 
        0x1,
        TraceFilterSimple(set((index,) + tuple(ignore)), trace_type),
        trace
    )

    return trace.end_position if trace.did_hit() else None

def geteyelocation(index):
    # Returns the eye location of a player

    # http://www.sourcepython.com/showthread.php?579-Player-View-Coordinates

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    return Vector(*tuple(player.get_prop_float(
        'm_vecViewOffset[{0}]'.format(x)) + y for x, y in \
            enumerate(player.location)))

def getviewvector(index):
    # Returns the view vector of a player as a mathlib.Vector instance

    # http://www.sourcepython.com/showthread.php?579-Player-View-Coordinates

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    eye_angle0 = player.get_prop_float('m_angEyeAngles[0]')
    eye_angle1 = player.get_prop_float('m_angEyeAngles[1]')

    return Vector(
        math.cos(math.radians(eye_angle1)),
        math.sin(math.radians(eye_angle1)),
        -1 * math.sin(math.radians(eye_angle0))
    )

def getcvarvalue(index, cvar):
    # Returns the player's value of the given cvar (e.g. cl_cmdrate)

    cvar = str(cvar)

    return engine_server.get_client_convar_value(index, cvar)

def getvelocity(index):
    # Returns the current velocity of a player

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    x = player.get_prop_float('m_vecVelocity[0]')
    y = player.get_prop_float('m_vecVelocity[1]')
    z = player.get_prop_float('m_vecVelocity[2]')
    return round((x*x + y*y + z*z)**0.5, 2)

def getspeed(index):
    # Returns the current speed multiplier of a player

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    return player.get_prop_float('m_flLaggedMovementValue')

def getnoclip(index):
    # Returns True if player is in noclip, otherwise False

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    return True if player.get_prop_int("movetype") == 8 else False

def getnoblock(index):
    # Returns True if player has noblock, otherwise False

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    return True if player.get_prop_int("m_CollisionGroup") == 2 else False

def getjetpack(index):
    # Returns True if player has jetpack, otherwise False

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    return True if player.get_prop_int("movetype") == 4 else False

def getfreeze(index):
    # Returns True if player is frozen, otherwise False

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    return True if player.get_prop_int("movetype") == 0 else False

def getgodmode(index):
    # Returns True if player is in godmode, otherwise False

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    return True if player.get_prop_int("m_lifeState") == 0 else False

def gethelmet(index):
    # Returns True if player has a helmet, otherwise False

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    return True if player.get_prop_int("m_bHasHelmet") == 1 else False

def getnearplayers(index, distance):
    # Returns a list of indexes of players within the given distance value

    playerlist = []
    distance = int(distance)

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    loc1 = player.location

    for id in PlayerIter('alive'):
        if id == index:
            continue

        if id not in _players:
            _players[id] = PlayerEntity(id)

        player2 = _players[id]
        loc2 = player2.location

        if loc1.get_distance(loc2) > distance:
            continue
        
        playerlist.append(id)

    return playerlist

def getclosestplayer(index, team=None):
    # Returns the index of the closest player of the given index

    playerlist = []

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    loc1 = player.location

    teams = {2:'t', 3:'ct'}

    if team:
        team = int(team)
        if team not in teams:
            raise(TypeError, 'Team must be 2 or 3')
        playeriter = PlayerIter('alive', teams[team])
    else:
        playeriter = PlayerIter('alive')

    for id in playeriter:
        if id == index:
            continue

        if id not in _players:
            _players[id] = PlayerEntity(id)

        player2 = _players[id]
        loc2 = player2.location

        playerlist.append((loc1.get_distance(loc2), id))

    if not playerlist:
        return (None, None)

    return min(playerlist)

def getweaponindex(index, weapon):
    # Get the weapon index by its name held by a player
    # Returns 0 if the player does not have the specified weapon

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    handle = player.inthandle

    for entid in EntityIter(weapon):
        ent = BaseEntity(entid)
        if handle == ent.get_prop_int("m_hOwnerEntity"):
            return entid

    return 0

def getweaponindexes(index):
    # Returns a list with all weapon indexes held by a player

    weapons = []

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    handle = player.inthandle

    for entid in EntityIter('weapon_', False):
        ent = BaseEntity(entid)
        if handle == ent.get_prop_int("m_hOwnerEntity"):
            weapons.append(entid)

    return weapons

def changeteam(index, team):
    # Changes the team of a player (kills the player)

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    player.set_team(team)

def sethealth(index, value):
    # Sets the health of a player

    value = int(value)

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    player.set_prop_int('m_iHealth', value)

def setspeed(index, value):
    # Sets the speed multiplier of a player

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    player.set_prop_float('m_flLaggedMovementValue', value)

def setmodel(index, model):
    # Set the model of a player

    model = str(model)
    if not model.startswith('models/'):
        model = "models/" + model

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    model_index = precachemodel(model)
    color = getcolor(index)
    player.set_prop_int('m_nModelIndex', model_index)
    setcolor(index, *color)

def setcolor(index, r, g, b, a=None):
    # Set the model color of a player

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    color = int(r) + (int(g) << 8) + (int(b) << 16)
    if a is None:
        color += getcolor(player)[3] << 24
    else:
        color += int(a) << 24
    if color >= 2**31: color -= 2**32
    player.set_prop_int("m_nRenderMode", player.get_prop_int("m_nRenderMode") | 1)
    player.set_prop_int("m_nRenderFX", player.get_prop_int("m_nRenderFX") | 256)
    player.set_prop_int("m_clrRender", color)

def setnoclip(index, value):
    # Toggles noclip on a player

    value = int(value)

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    if value == 1:
        player.set_prop_int("movetype", 8)
    elif value == 0:
        player.set_prop_int("movetype", 2)
    else:
        raise(ValueError, 'Invalid value for noclip')

def setnoblock(index, value):
    # Toggles noblock on a player

    value = int(value)

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    if value == 1:
        player.set_prop_int("m_CollisionGroup", 2)
    elif value == 0:
        player.set_prop_int("m_CollisionGroup", 5)
    else:
        raise(ValueError, 'Invalid value for noblock')

def setjetpack(index, value):
    # Toggles jetpack on a player

    value = int(value)

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    if value == 1:
        player.set_prop_int("movetype", 4)
    elif value == 0:
        player.set_prop_int("movetype", 2)
    else:
        raise(ValueError, 'Invalid value for jetpack')

def setfreeze(index, value):
    # Toggles freeze on a player

    value = int(value)

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    if value == 1:
        player.set_prop_int("movetype", 0)
    elif value == 0:
        player.set_prop_int("movetype", 2)
    else:
        raise(ValueError, 'Invalid value for freeze')

def godmode(index, value):
    # Toggles godmode on a player

    value = int(value)

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]
    if value == 1:
        player.set_prop_int("m_lifeState", 0)
    elif value == 0:
        player.set_prop_int("m_lifeState", 512)
    else:
        raise(ValueError, 'Invalid value for godmode')

def flash(index, alpha, duration):
    # Blinds a player

    alpha = int(alpha)
    duration = int(duration)

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    player.set_prop_float("m_flFlashMaxAlpha", 0)
    player.set_prop_float("m_flFlashDuration", 0)
    player.set_prop_float("m_flFlashMaxAlpha", alpha)
    player.set_prop_float("m_flFlashDuration", duration)

def blind(index, alpha, duration):
    # Blinds player

    flash(index, alpha, duration)

def sethelmet(index, value):
    # Gives the player a helmet or removes it

    value = int(value)

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    player.set_prop_int("m_bHasHelmet", value)

def getplayercount():
    # Returns number of players and bots

    x = 0
    for index in PlayerIter():
        x += 1

    return x

def gethumancount():
    # Returns number of players

    x = 0
    for index in PlayerIter('human'):
        x += 1

    return x

def getbotcount():
    # Returns number of bots

    x = 0
    for index in PlayerIter(not_filters='human'):
        x += 1

    return x

def getalivecount():
    # Returns the number of alive players and bots

    x = 0
    for index in PlayerIter('alive'):
        x += 1

    return x

def getdeadcount():
    # Returns the number of dead players and bots

    x = 0
    for index in PlayerIter('dead'):
        x += 1

    return x

def precachemodel(model):
    # Precaches a model and returns its index

    if model not in _precached_models:
        _precached_models[model] = engine_server.precache_model(model)

    return _precached_models[model]

def precachedecal(decal):
    # Precaches a decal and returns its index

    if decal not in _precached_models:
        _precached_models[decal] = engine_server.precache_decal(decal)

    return _precached_models[decal]

def clientcommand(index, command):
    # Forces a player to execute a command

    if index not in _players:
        _players[index] = PlayerEntity(index)

    player = _players[index]

    engine_server.client_command(player.edict, command)

def createbot(name):
    # Creates a fake client with the given name

    name = str(name)

    engine_server.create_fake_client(name)

def drawline(start, end, model="sprites/laserbeam.vmt", halo="sprites/halo.vmt", startframe=0, framerate=1, life=1, width=5, endwidth=5, fadelength=5, amplitude=0, red=255, green=0, blue=0, alpha=255, speed=0):
    # Draw a line for every player

    model_index = precachemodel(model)
    halo_index = precachemodel(halo)

    for index in PlayerIter('human'):
        temp_entities.beam_points(RecipientFilter(index), 0, start, end, model_index, halo_index, startframe, framerate,
            life, width, endwidth, fadelength, amplitude, 
            red, green, blue, alpha, speed
        )

def drawcircle(center, start_radius=0, end_radius=250, model="sprites/laserbeam.vmt", halo="sprites/halo.vmt", startframe=0, framerate=255, life=1, width=5, spread=0, amplitude=0, red=255, green=0, blue=0, alpha=255, speed=1, flags=0):
    # Draw a circle to every player    

    model_index = precachemodel(model)
    halo_index = precachemodel(halo)

    for index in PlayerIter('human'):
        temp_entities.beam_ring_point(RecipientFilter(index), 0, center, start_radius, end_radius, model_index, halo_index, startframe, framerate,
            life, width, spread, amplitude, 
            red, green, blue, alpha, speed, flags
        )

def tell(index, message):
    # Send a message to a certain player
    SayText2(message=message).send(index)

def msg(message):
    # Send a message to every player
    SayText2(message=message).send()

def hudhint(index, message):
    # Send a HintText to a certain player
    HintText(message=message).send(index)